"""
check coco dataset
"""
import json
import random
import os
import argparse
from pathlib import Path
import cv2
import warnings

from ..functions import general
from ..controls.color import ColorControl


class CheckCOCO(object):
	def __init__(self, json_path, img_dir=None, line_thickness=3):
		self.jp = json_path  # json annotation path
		p = Path(self.jp)
		self.img_dir = img_dir if img_dir is not None else f'{p.parent.parent}/{str(p.stem).split("_")[-1]}'
		self.line_thickness = line_thickness

	def __call__(self, only_show_id=None):
		print(f'Loading annotations...')
		with open(self.jp, 'r') as f:
			annoj = json.load(f)

		print(f'\n{"":=<30} Dataset Info {"":=<30}')
		print(f'The annotation_json contains {len(annoj)} keys: {list(annoj.keys())}')  # 统计json文件的关键字长度
		# info
		info = annoj['info']
		print(f'Dataset info: {info}')
		licenses = annoj['licenses']
		print(f'Dataset license: {licenses[0]}')

		# 关于类别
		classes = annoj['categories']
		classes = {x['id']: x['name'] for x in classes}
		print(f'The dataset contains {len(classes)} classes: {classes}')
		colors = ColorControl(num=len(classes), rgb='BGR').color

		# 关于images
		images = annoj['images']
		print(f'The dataset contains {len(images)} images. ', end='')  # json文件中包含的图片数量
		print(f'Image example: {images[0]}')

		# 关于annos
		annos = annoj['annotations']
		print(f'The dataset contains {len(annos)} targets. ', end='')
		print(f'Target example: {annos[0]}')

		print(f'\n{"":=<30}  Image Info  {"":=<30}')
		# 可视化图像
		annos_img_ids = [x['image_id'] for x in annos]
		# print(annos_img_ids, len(annos_img_ids))
		for i, img_info in enumerate(images):
			img_name = img_info['file_name']
			img_id = img_info['id']
			height, width = img_info['height'], img_info['width']
			date_captured = img_info['date_captured']

			# 找到对应图像
			img_path = f'{self.img_dir}/{img_name}'
			assert os.path.exists(img_path), f'img {img_path} does not exists.'
			img = cv2.imread(img_path)
			h, w, c = img.shape
			assert height == h and width == w, f'The shape of image: {h}*{w} is inconsistent with that from ' \
											   f'annotation: {height}*{width}. img_path: {img_path}'

			# 找对应的标注
			# anno_idx = annos_img_ids.index(img_id)  # 这是错的，一张图可能对应多个annos，这种方法只会找到一个
			anno_idxes = [j for j, x in enumerate(annos_img_ids) if img_id == x]

			if len(anno_idxes) == 0:
				print(f'Image {img_name} without targets.')

			img_annos = [annos[idx] for idx in anno_idxes]
			num_t = len(img_annos)
			# print(img_name, img_annos)

			for anno in img_annos:
				seg = anno['segmentation']
				bbox = anno['bbox']  # x1 y1 w h, in pixels
				bbox[2] += bbox[0]
				bbox[3] += bbox[1]  # 转为x1 y1 x2 y2
				anno_img_id = anno['image_id']
				cls = anno['category_id']  # 1 2 3 ...
				cls_name = classes[cls]
				area = anno['area']
				# print(anno)

				color_idx = list(classes.values()).index(cls_name)
				# print(f'color: {}')
				img = general.plot_one_box_trace_pose_status(
					bbox, img, color=colors[color_idx], label=cls_name,
					line_thickness=self.line_thickness)

			print(f'img_id: {img_id:<6} img_name: {img_name} num_targets: {num_t}')

			if only_show_id is None or str(img_id) == str(only_show_id):
				cv2.imshow(f'{img_name}', img)
				if cv2.waitKey(0) == ord('q'):
					break
				cv2.destroyAllWindows()


if __name__ == "__main__":
	pass



import os, sys
import cv2
from .check_coco import CheckCOCO
from .check_yolo import CheckYOLO
from pathlib import Path
import shutil
import warnings


class Tools(object):
	def __init__(self):
		pass

	def test(self):
		print('z')

	def cap_video_save(self, path, new_size=None, rotate=False, save_dir=None):
		cap = cv2.VideoCapture(path)
		assert cap.isOpened(), f'Failed to open {path}'
		w = int(cap.get(cv2.CAP_PROP_FRAME_WIDTH))
		h = int(cap.get(cv2.CAP_PROP_FRAME_HEIGHT))
		fps = cap.get(cv2.CAP_PROP_FPS) % 100
		print(f'{path}, w: {w}, h: {h}, fps: {fps}')
		record_flag = False
		idx = 0
		while True:
			ret, frame = cap.read()
			if rotate:
				frame = cv2.rotate(frame, cv2.ROTATE_180)
			if new_size:
				frame = cv2.resize(frame, new_size, interpolation=cv2.INTER_CUBIC)
			cv2.imshow('x', frame)
			if cv2.waitKeyEx(1) == ord('q'):
				cap.release()
				break
			if cv2.waitKeyEx(1) == ord('r'):
				record_flag = True
			if record_flag and save_dir is not None:
				print(f'\rrecording ... ', end='')
				cv2.imwrite(f'{save_dir}/{idx:0>6}.jpg', frame)

			idx += 1
			print(f' {idx:0>6}.jpg')

	def yolo2coco(self, sp, tp=None):
		"""
		transform datatset of yolov5 format to coco format
		:param sp: source path, path to yolov5 dataset dir, which contains images, label dir and classes.txt
		:param sp: target path, output path
		"""
		from .yolo2coco import YOLO2COCO
		# print(f'sp: {sp}\ntp: {tp}')
		y2c = YOLO2COCO(sp, tp=tp)
		y2c()

	def check_COCO(self, json_path, img_dir=None, line_thicknes=3, only_show_id=None):
		"""check coco dataset
		:param json_path, path to coco json file, for example: /xxx/annotations/instances_train2017.json
		:param img_dir, image dir
		:param line_thicknes: thickness for rectangle of bbox
		:param only_show_id: only show the img with this id if is None
		"""
		cc = CheckCOCO(json_path, img_dir=img_dir, line_thickness=line_thicknes)
		cc(only_show_id=only_show_id)

	def check_YOLO(self, dp, trte=None):
		"""
		check YOLO dataset
		:param dp: yolo dataset dir, for example: /xxx/xxx_yolo_format
		:return: None
		"""
		cy = CheckYOLO(dp=dp)
		cy(trte=trte)


def video2frames(video_path, output_path=None, decoder='ffmpeg', *args, **kwargs):
	""""""
	output_path = output_path if output_path else f'{Path(video_path).parent}/{Path(video_path).stem}'
	interval = kwargs.pop('interval', 5)
	digits = kwargs.pop('digits', 6)

	# print(output_path)

	if os.path.exists(output_path):
		shutil.rmtree(output_path)
	os.makedirs(output_path)

	if decoder == 'ffmpeg':
		pass
		code = \
			fr'ffmpeg -i {video_path} -vf "select=not(mod(n\, {interval}))" ' \
			fr'-y -acodec copy -vsync 0 {output_path}/%0{digits}d.jpg'
		print(f'exec code: {code}')
		os.system(code)
	elif decoder == 'cv2' or decoder == 'opencv':
		# print "正在读取视频：", each_video
		print("正在读取视频：", video_path)  # 我的是Python3.6

		cap = cv2.VideoCapture(video_path)
		frame_index = 0
		frame_count = 0
		if cap.isOpened():
			success = True
		else:
			success = False
			warnings.warn(f'read video: {video_path} failed')

		while (success):
			success, frame = cap.read()
			# print "---> 正在读取第%d帧:" % frame_index, success
			print(f"\r读取第{frame_index}帧, 保存第{frame_count}帧: {success}", end='')

			if frame_index % interval == 0 and success:  # 如路径下有多个视频文件时视频最后一帧报错因此条件语句中加and success
				# resize_frame = cv2.resize(frame, (frame_width, frame_height), interpolation=cv2.INTER_AREA)
				# cv2.imwrite(each_video_save_full_path + each_video_name + "_%d.jpg" % frame_index, resize_frame)
				img_save_path = f"{output_path}/{frame_count:0>{digits}}.jpg"
				# cv2.imwrite(each_video_save_full_path + "%d.jpg" % frame_count, resize_frame)
				cv2.imwrite(img_save_path, frame)
				frame_count += 1
			frame_index += 1

		cap.release()  # 这行要缩一下、原博客会报错
	else:
		raise NotImplementedError(f'decoder: {decoder}')



"""
dm关于时间的函数
"""
import datetime


def current_time(format='%Y-%m-%d %H:%M:%S'):
	"""
	读取当前时间
	:param format: format
	:return: time, str
	"""
	return datetime.datetime.now().strftime(format)


def plus_time(ct=None, seconds=1, format='%Y-%m-%d %H:%M:%S'):
	"""
	添加时间
	:param ct: current time, str
	:param seconds: seconds, int
	:param format: format, str
	:return: time, str
	"""
	ct = ct if ct else current_time(format=format)
	ct = datetime.datetime.strptime(ct, format)
	return (ct + datetime.timedelta(seconds=seconds)).strftime(format)


def within_time(expiration_time, current_t=None, format='%Y-%m-%d %H:%M:%S'):
	"""
	输入到期时间，判断当前时间是否在到期时效内
	:param expiration_time: 到期时间, str
	:param current_t: 当前时间，str
	:param format: 格式，str
	:return: True or False, bool
	"""
	ct = current_t if current_t else current_time(format=format)  # 读取当前时间，str格式
	# ct = plus_time(ct, seconds=50*24*60*60)  # 测试用的，
	ct = datetime.datetime.strptime(ct, format)  # 转为date格式
	et = datetime.datetime.strptime(expiration_time, format)  # 转为date格式
	return ct < et



__version__ = '1.0.165'
__author__ = 'zhengzhengde  mail: drivener@163.com'

"""
更新日志：
20200808 v1.0.0 初始版本

20210303 v1.0.15 general.letterbox
				general.plot_one_box_trace_pose_status
				general.xyxy2xywh
				general.xywh2xyxy
				dm.
20211010 v1.0.153 dm.tools.check_COCO dm.tools.check_YOLO
20211201 v1.0.154 数据增强，根据mask的目标的随机位置、随机缩放和随机旋转，还没写好啊
20211210 v1.0.155 dm.genral.pts2bbox 
20211223 v1.0.156 dm.video2frames
20211223 v1.0.157 dm.misc.list2table dm.misc.flatten_list
20220114 v1.0.158 dm.general.torch_distributed_zero_first
20220215 v1.0.160

"""



"""
函数集合
"""
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.pyplot import MultipleLocator


def ReLU(x):
	x = 1 * (x > 0) * x
	return x


def FastPower(x, power=3):
	x = ReLU(x)
	return np.power(x + (1 - x) / 2, power)


def ImDU(x, power=3):
	return ReLU(1 - FastPower(x, power=power))


if __name__ == '__main__':
	#
	ax = plt.gca()
	ax.spines['right'].set_color('none')
	ax.spines['top'].set_color('none')
	ax.xaxis.set_ticks_position('bottom')
	ax.yaxis.set_ticks_position('left')
	ax.spines['bottom'].set_position(('data', 0))
	ax.spines['left'].set_position(('data', 0))
	ax.xaxis.set_major_locator(MultipleLocator(0.5))
	ax.yaxis.set_major_locator(MultipleLocator(0.5))

	x = np.linspace(0, 2, 100)
	print(x)
	f = FastPower
	# f = ImDU
	y = f(x, power=8)
	y = ImDU(x, 8)
	print(y)
	color = '#EE3B3B'
	# plt.plot((-1, 0), (0, 0), color=color, linewidth=5)
	# plt.plot((0, 0), (0, 1), color=color, linewidth=5)
	plt.plot(x, y, color=color, linewidth=5)
	# plt.plot(x, f(x, power=3))
	plt.xlim(-0.2, 1.2)
	plt.ylim(-0.2, 1.2)

	fontdict = {'size': 18, 'weight': 'bold'}
	# plt.legend(prop=fontdict)
	# plt.ylabel(f'1 (eV)', fontdict=fontdict)
	plt.yticks(size=16, weight='bold')
	plt.xticks(size=16, weight='bold')
	plt.show()



"""
functions of dm
"""
import math
import numpy as np
import cv2
import random
from contextlib import contextmanager

try:
	import torch
except Exception as e:
	pass

def letterbox(
		img, new_shape=640, color=(114, 114, 114), auto=True, scaleFill=False, scaleup=True, roi=None):
	"""
	resize image to new_shape with multiple of 32
	:param img: np.array [w, h, 3]
	:param new_shape: int or tuple, 640, [640, 320]
	:param color: background color
	:param auto: auto pad
	:param scaleFill: stretch pad
	:param scaleup: scale up and scale down if Ture else only scale down
	:param roi: use roi or not
	:return: img, ratio, (dw, dh), recover
	"""
	# Resize image to a 32-pixel-multiple rectangle https://github.com/ultralytics/yolov3/issues/232
	shape = img.shape[:2]  # current shape [height, width]  720, 1280

	if isinstance(new_shape, int):
		new_shape = (new_shape, new_shape)

	# Scale ratio (new / old)
	r = min(new_shape[0] / shape[0], new_shape[1] / shape[1])
	if not scaleup:  # only scale down, do not scale up (for better tests mAP)
		r = min(r, 1.0)

	# Compute padding
	ratio = r, r  # width, height ratios
	new_unpad = int(round(shape[1] * r)), int(round(shape[0] * r))
	dw, dh = new_shape[1] - new_unpad[0], new_shape[0] - new_unpad[1]  # wh padding
	# print(dw, dh, ratio)
	if auto:  # minimum rectangle
		dw, dh = np.mod(dw, 64), np.mod(dh, 64)  # wh padding
	elif scaleFill:  # stretch
		dw, dh = 0.0, 0.0
		new_unpad = (new_shape[1], new_shape[0])
		ratio = new_shape[1] / shape[1], new_shape[0] / shape[0]  # width, height ratios

	# print(dw, dh)
	dw /= 2  # divide padding into 2 sides
	dh /= 2
	# print(new_unpad)

	if shape[::-1] != new_unpad:  # resize
		img = cv2.resize(img, new_unpad, interpolation=cv2.INTER_LINEAR)

	# roi
	if roi is not None:
		x1, x2 = roi[0] / shape[1] * new_unpad[0], roi[2] / shape[1] * new_unpad[0]  # convert from pixel to percet
		y1, y2 = roi[1] / shape[0] * new_unpad[1], roi[3] / shape[0] * new_unpad[1]
		img = img[int(y1):int(y2), int(x1):int(x2)]
		rest_h = img.shape[0] % 32
		rest_w = img.shape[1] % 32
		dh = 0 if rest_h == 0 else (32 - rest_h) / 2
		dw = 0 if rest_w == 0 else (32 - rest_w) / 2
		recover = [new_shape[0], new_unpad[1], int(x1) - dw, int(y1) - dh]
	else:
		recover = None

	top, bottom = int(round(dh - 0.1)), int(round(dh + 0.1))
	left, right = int(round(dw - 0.1)), int(round(dw + 0.1))
	img = cv2.copyMakeBorder(img, top, bottom, left, right, cv2.BORDER_CONSTANT, value=color)  # add border
	return img, ratio, (dw, dh), recover


def plot_one_box_trace_pose_status(
		x, img, color=None, label=None, line_thickness=None, focus=False, trace=None, status=None,
		keypoints=None, kp_score=None):
	"""
	draw box in img, support bbox, trace, pose and status, and focus
	:param x: bbox in xyxy format
	:param img: raw_img
	:param color: color in (R, G, B)
	:param label: label for target detection, i.e. cls or tid
	:param line_thickness: rt
	:param focus: is fill the inner bbox area
	:param trace: trace
	:param status: target status
	:param keypoints: tuple, [num_kps, 2], keypoints in pixel, support num_kps: 17 26 136
	:param kp_score: tuple, [num_kps, 1], score of every keypoint
	:return:
	"""
	# Plots one bounding box on image img
	tl = line_thickness or round(0.002 * (img.shape[0] + img.shape[1]) / 2) + 1  # line/font thickness
	color = color or [random.randint(0, 255) for _ in range(3)]
	c1, c2 = (int(x[0]), int(x[1])), (int(x[2]), int(x[3]))
	cv2.rectangle(img, c1, c2, color, thickness=tl, lineType=cv2.LINE_AA)
	# print('plot', img.shape, c1, c2, color, tl)
	if focus:  # 把框內的部分也填入
		focus_area = np.zeros((c2[1] - c1[1], c2[0] - c1[0], 3), dtype=np.uint8)
		focus_area[:, :, 0::] = color
		img = imgAdd(focus_area, img, x=c1[1], y=c1[0], alpha=0.75)

	if trace is not None:  # 如果传入了轨迹，就绘制轨迹，传入的轨迹是nparray (N, 2) N>=2，后面的2的xy
		# print('trace', trace)
		for i in range(len(trace) - 1):
			pt1 = tuple(trace[i])
			# pt2 = tuple(trace[i] + 1)  # 之前写的用在AILabelImage上的，没有报错啊
			pt2 = tuple(trace[i + 1])
			cv2.arrowedLine(img, pt1, pt2, color, int(2 * tl), cv2.LINE_8, 0, 0.3)

	if label:
		tf = max(tl - 1, 1)  # font thickness
		t_size = cv2.getTextSize(label, 0, fontScale=tl / 3, thickness=tf)[0]
		c2 = c1[0] + t_size[0], c1[1] - t_size[1] - 3
		cv2.rectangle(img, c1, c2, color, -1, cv2.LINE_AA)  # filled
		cv2.putText(img, label, (c1[0], c1[1] - 2), 0, tl / 3, [225, 255, 255], thickness=tf, lineType=cv2.LINE_AA)
		if status is not None:  # 绘制状态，字体稍微小一点
			status = [status] if isinstance(status, str) else status
			# print(status)
			assert len(status) <= 2  # 最多两个状态类型，也可以是1个
			# print('s')
			spt1 = (int(x[2]), int(x[1]))  # status point1
			if len(status) == 1:
				max_status = status[0]
			else:
				max_status = status[0] if len(status[0]) >= len(status[1]) else status[1]  # 选择pose和action中最长的那个来计算size

			s_size = cv2.getTextSize(max_status, 0, fontScale=tl / 4, thickness=tf)[0]
			spt2 = (spt1[0] + s_size[0], spt1[1] + len(status) * s_size[1] + 15)
			cv2.rectangle(img, spt1, spt2, color, -1, cv2.LINE_AA)

			cv2.putText(
				img, status[0], (spt1[0], spt1[1] + s_size[1] + 4), 0, tl / 4, [255, 255, 255], thickness=tf,
				lineType=cv2.LINE_AA)
			if len(status) == 2:
				cv2.putText(
					img, status[1], (spt1[0], spt1[1] + 2 * s_size[1] + 10), 0, tl / 4, [255, 255, 255], thickness=tf,
					lineType=cv2.LINE_AA)
	# print(status, t_size, s_size, spt1, spt2, tl)

	if keypoints is not None:
		kp_preds = np.array(keypoints)
		kp_scores = np.array(kp_score)
		kp_num = len(kp_preds)

		if kp_num == 17:
			kpformat = 'coco'
			if kpformat == 'coco':
				l_pair = [
					(0, 1), (0, 2), (1, 3), (2, 4),  # Head
					(5, 6), (5, 7), (7, 9), (6, 8), (8, 10),
					(17, 11), (17, 12),  # Body
					(11, 13), (12, 14), (13, 15), (14, 16)
				]
				p_color = [
					(0, 255, 255), (0, 191, 255), (0, 255, 102), (0, 77, 255), (0, 255, 0),
					# Nose, LEye, REye, LEar, REar
					(77, 255, 255), (77, 255, 204), (77, 204, 255), (191, 255, 77), (77, 191, 255),
					(191, 255, 77),  # LShoulder, RShoulder, LElbow, RElbow, LWrist, RWrist
					(204, 77, 255), (77, 255, 204), (191, 77, 255), (77, 255, 191), (127, 77, 255),
					(77, 255, 127), (0, 255, 255)]  # LHip, RHip, LKnee, Rknee, LAnkle, RAnkle, Neck
				line_color = [
					(0, 215, 255), (0, 255, 204), (0, 134, 255), (0, 255, 50),
					(77, 255, 222), (77, 196, 255), (77, 135, 255), (191, 255, 77), (77, 255, 77),
					(77, 222, 255), (255, 156, 127),
					(0, 127, 255), (255, 127, 77), (0, 77, 255), (255, 77, 36)]
			else:
				raise NotImplementedError
		elif kp_num == 136:
			l_pair = [
				(0, 1), (0, 2), (1, 3), (2, 4),  # Head
				(5, 18), (6, 18), (5, 7), (7, 9), (6, 8), (8, 10),  # Body
				(17, 18), (18, 19), (19, 11), (19, 12),
				(11, 13), (12, 14), (13, 15), (14, 16),
				(20, 24), (21, 25), (23, 25), (22, 24), (15, 24), (16, 25),  # Foot
				(26, 27), (27, 28), (28, 29), (29, 30), (30, 31), (31, 32), (32, 33), (33, 34), (34, 35), (35, 36),
				(36, 37), (37, 38),  # Face
				(38, 39), (39, 40), (40, 41), (41, 42), (43, 44), (44, 45), (45, 46), (46, 47), (48, 49), (49, 50),
				(50, 51), (51, 52),  # Face
				(53, 54), (54, 55), (55, 56), (57, 58), (58, 59), (59, 60), (60, 61), (62, 63), (63, 64), (64, 65),
				(65, 66), (66, 67),  # Face
				(68, 69), (69, 70), (70, 71), (71, 72), (72, 73), (74, 75), (75, 76), (76, 77), (77, 78), (78, 79),
				(79, 80), (80, 81),  # Face
				(81, 82), (82, 83), (83, 84), (84, 85), (85, 86), (86, 87), (87, 88), (88, 89), (89, 90), (90, 91),
				(91, 92), (92, 93),  # Face
				(94, 95), (95, 96), (96, 97), (97, 98), (94, 99), (99, 100), (100, 101), (101, 102), (94, 103),
				(103, 104), (104, 105),  # LeftHand
				(105, 106), (94, 107), (107, 108), (108, 109), (109, 110), (94, 111), (111, 112), (112, 113),
				(113, 114),  # LeftHand
				(115, 116), (116, 117), (117, 118), (118, 119), (115, 120), (120, 121), (121, 122), (122, 123),
				(115, 124), (124, 125),  # RightHand
				(125, 126), (126, 127), (115, 128), (128, 129), (129, 130), (130, 131), (115, 132), (132, 133),
				(133, 134), (134, 135)  # RightHand
			]
			p_color = [
				(0, 255, 255), (0, 191, 255), (0, 255, 102), (0, 77, 255), (0, 255, 0),  # Nose, LEye, REye, LEar, REar
				(77, 255, 255), (77, 255, 204), (77, 204, 255), (191, 255, 77), (77, 191, 255),
				(191, 255, 77),  # LShoulder, RShoulder, LElbow, RElbow, LWrist, RWrist
				(204, 77, 255), (77, 255, 204), (191, 77, 255), (77, 255, 191), (127, 77, 255),
				(77, 255, 127),  # LHip, RHip, LKnee, Rknee, LAnkle, RAnkle, Neck
				(77, 255, 255), (0, 255, 255), (77, 204, 255),  # head, neck, shoulder
				(0, 255, 255), (0, 191, 255), (0, 255, 102), (0, 77, 255), (0, 255, 0),
				(77, 255, 255)]  # foot

			line_color = [
				(0, 215, 255), (0, 255, 204), (0, 134, 255), (0, 255, 50),
				(0, 255, 102), (77, 255, 222), (77, 196, 255), (77, 135, 255), (191, 255, 77),
				(77, 255, 77),
				(77, 191, 255), (204, 77, 255), (77, 222, 255), (255, 156, 127),
				(0, 127, 255), (255, 127, 77), (0, 77, 255), (255, 77, 36),
				(0, 77, 255), (0, 77, 255), (0, 77, 255), (0, 77, 255), (255, 156, 127), (255, 156, 127)]
		elif kp_num == 26:
			l_pair = [
				(0, 1), (0, 2), (1, 3), (2, 4),  # Head
				(5, 18), (6, 18), (5, 7), (7, 9), (6, 8), (8, 10),  # Body
				(17, 18), (18, 19), (19, 11), (19, 12),
				(11, 13), (12, 14), (13, 15), (14, 16),
				(20, 24), (21, 25), (23, 25), (22, 24), (15, 24), (16, 25),  # Foot
			]
			p_color = [
				(0, 255, 255), (0, 191, 255), (0, 255, 102), (0, 77, 255), (0, 255, 0),  # Nose, LEye, REye, LEar, REar
				(77, 255, 255), (77, 255, 204), (77, 204, 255), (191, 255, 77), (77, 191, 255),
				(191, 255, 77),  # LShoulder, RShoulder, LElbow, RElbow, LWrist, RWrist
				(204, 77, 255), (77, 255, 204), (191, 77, 255), (77, 255, 191), (127, 77, 255),
				(77, 255, 127),  # LHip, RHip, LKnee, Rknee, LAnkle, RAnkle, Neck
				(77, 255, 255), (0, 255, 255), (77, 204, 255),  # head, neck, shoulder
				(0, 255, 255), (0, 191, 255), (0, 255, 102), (0, 77, 255), (0, 255, 0),
				(77, 255, 255)]  # foot

			line_color = [
				(0, 215, 255), (0, 255, 204), (0, 134, 255), (0, 255, 50),
				(0, 255, 102), (77, 255, 222), (77, 196, 255), (77, 135, 255), (191, 255, 77),
				(77, 255, 77),
				(77, 191, 255), (204, 77, 255), (77, 222, 255), (255, 156, 127),
				(0, 127, 255), (255, 127, 77), (0, 77, 255), (255, 77, 36),
				(0, 77, 255), (0, 77, 255), (0, 77, 255), (0, 77, 255), (255, 156, 127), (255, 156, 127)]
		else:
			raise NotImplementedError

		# draw keypoints
		vis_thres = 0.05 if kp_num == 136 else 0.4
		part_line = {}
		for n in range(kp_scores.shape[0]):
			if kp_scores[n] <= vis_thres:
				continue
			cor_x, cor_y = int(kp_preds[n, 0]), int(kp_preds[n, 1])
			part_line[n] = (cor_x, cor_y)
			if n < len(p_color):
				cv2.circle(img, (cor_x, cor_y), 3, p_color[n], -1)
			else:
				cv2.circle(img, (cor_x, cor_y), 1, (255, 255, 255), 2)
		# Draw limbs
		for i, (start_p, end_p) in enumerate(l_pair):
			if start_p in part_line and end_p in part_line:
				start_xy = part_line[start_p]
				end_xy = part_line[end_p]
				if i < len(line_color):
					cv2.line(img, start_xy, end_xy, line_color[i], 2 * int(kp_scores[start_p] + kp_scores[end_p]) + 1)
				else:
					cv2.line(img, start_xy, end_xy, (255, 255, 255), 1)
	return img


def imgAdd(small_img, big_image, x, y, alpha=0.5):
	"""
	draw small image into big image
	:param small_img: small img
	:param big_image: big img
	:param x: left position for drawing in pixel
	:param y: top position for drawing in pixel
	:param alpha: transparency
	:return: big img draw with small img
	"""
	row, col = small_img.shape[:2]
	if small_img.shape[0] > big_image.shape[0] or small_img.shape[1] > big_image.shape[1]:
		raise NameError(f'imgAdd, the size of small img bigger than big img.')
	roi = big_image[x:x + row, y:y + col, :]
	roi = cv2.addWeighted(small_img, alpha, roi, 1 - alpha, 0)
	big_image[x:x + row, y:y + col] = roi
	return big_image


def imgMask(img1, img2, mask):
	"""
	mask为1的像素取img1, mask为0的像素取img2
	:param img1: image 1 ndarray [h, w, c]
	:param img2: image 2 ndarray [h, w, c]
	:param mask: mask ndarray [h, w]
	:return:
	"""
	pass


def bbox_iou(box1, box2, x1y1x2y2=True, GIoU=False, DIoU=False, CIoU=False, return_np=False):
	"""
	input box1 and box2, return the iou.
	:param box1: torch.tensor or numpy.array, shape: (4,) normlized
	:param box2: torch.tensor or numpy.array, shape: (4,) normlized
	:param x1y1x2y2: True, if True, treat box as x1y1x2y2 format else xcycwh format
	:param GIoU: False, if True, cal GIoU
	:param DIoU: False, if True, cal DIou
	:param CIoU: False, if True, cal CIou
	:param return_np: Fasle, if True, return numpy.array type, else torch.tensor type
	:return: the IoU or GIoU or DIou OR CIoU of box1 and box2
	"""

	if isinstance(box1, np.ndarray):
		box1 = torch.from_numpy(box1)
	if isinstance(box2, np.ndarray):
		box2 = torch.from_numpy(box2)

	box2 = box2.T
	# Get the coordinates of bounding boxes
	if x1y1x2y2:  # x1, y1, x2, y2 = box1
		b1_x1, b1_y1, b1_x2, b1_y2 = box1[0], box1[1], box1[2], box1[3]
		b2_x1, b2_y1, b2_x2, b2_y2 = box2[0], box2[1], box2[2], box2[3]
	else:  # transform from xywh to xyxy
		b1_x1, b1_x2 = box1[0] - box1[2] / 2, box1[0] + box1[2] / 2
		b1_y1, b1_y2 = box1[1] - box1[3] / 2, box1[1] + box1[3] / 2
		b2_x1, b2_x2 = box2[0] - box2[2] / 2, box2[0] + box2[2] / 2
		b2_y1, b2_y2 = box2[1] - box2[3] / 2, box2[1] + box2[3] / 2

	# Intersection area
	inter = (torch.min(b1_x2, b2_x2) - torch.max(b1_x1, b2_x1)).clamp(0) * \
			(torch.min(b1_y2, b2_y2) - torch.max(b1_y1, b2_y1)).clamp(0)

	# Union Area
	w1, h1 = b1_x2 - b1_x1, b1_y2 - b1_y1
	w2, h2 = b2_x2 - b2_x1, b2_y2 - b2_y1
	union = (w1 * h1 + 1e-16) + w2 * h2 - inter

	# print(inter, union)
	iou = inter / union  # iou
	if GIoU or DIoU or CIoU:
		cw = torch.max(b1_x2, b2_x2) - torch.min(b1_x1, b2_x1)  # convex (smallest enclosing box) width
		ch = torch.max(b1_y2, b2_y2) - torch.min(b1_y1, b2_y1)  # convex height
		if GIoU:  # Generalized IoU https://arxiv.org/pdf/1902.09630.pdf
			c_area = cw * ch + 1e-16  # convex area
			return iou - (c_area - union) / c_area  # GIoU
		if DIoU or CIoU:  # Distance or Complete IoU https://arxiv.org/abs/1911.08287v1
			# convex diagonal squared
			c2 = cw ** 2 + ch ** 2 + 1e-16
			# centerpoint distance squared
			rho2 = ((b2_x1 + b2_x2) - (b1_x1 + b1_x2)) ** 2 / 4 + ((b2_y1 + b2_y2) - (b1_y1 + b1_y2)) ** 2 / 4
			if DIoU:
				return iou - rho2 / c2  # DIoU
			elif CIoU:  # https://github.com/Zzh-tju/DIoU-SSD-pytorch/blob/master/utils/box/box_utils.py#L47
				v = (4 / math.pi ** 2) * torch.pow(torch.atan(w2 / h2) - torch.atan(w1 / h1), 2)
				with torch.no_grad():
					alpha = v / (1 - iou + v + 1e-16)
				return iou - (rho2 / c2 + v * alpha)  # CIoU
	if return_np:
		iou = iou.numpy()
	# print(box1, box2, box1.shape, box2.shape, iou)
	return iou


def confusion2score(cm, round=5):
	"""
	input n*n confusion matrix, output: P R F1 score of each class and average ACC
	:param cm: confusion matrix, np.array, (n, n)
	:param round: default 5, decimal places
	:return: P, R, F1 and ACC
		P, R, F1: np.array (n,)
		acc: float32
	"""
	confusion = cm
	assert confusion.shape[0] == confusion.shape[1]
	nc = confusion.shape[0]
	P_sum = np.repeat(np.sum(confusion, axis=1) + 1e-10, nc).reshape(nc, nc)
	P = confusion / P_sum
	R_sum = np.repeat(np.sum(confusion, axis=0) + 1e-10, nc).reshape(nc, nc).transpose()
	R = confusion / R_sum

	# 取对角
	P = np.diagonal(P, offset=0)
	R = np.diagonal(R, offset=0)
	F1 = 2 * P * R / (P + R + 1e-10)

	# acc
	correct = np.diagonal(confusion, offset=0)
	acc = np.sum(correct) / (np.sum(confusion) + 1e-10)
	# print(confusion, P, R, F1, acc)
	if isinstance(round, int):
		P = np.round(P, round)
		R = np.round(R, round)
		F1 = np.round(F1, round)
		acc = np.round(acc, round)
	return P, R, F1, acc


@contextmanager
def torch_distributed_zero_first(local_rank: int):
	"""
	Decorator to make all processes in distributed training wait for each local_master to do something.
	"""
	if local_rank not in [-1, 0]:
		torch.distributed.barrier()
	yield
	if local_rank == 0:
		torch.distributed.barrier()

def xyxy2xywh(x):
	"""
	既要支持多个bbox转，也要支持单个bbox转
	:param x: bbox in [x1, y1, x2, y2] format of torch.tensor or np.array.
	:return: bbox in [xc, yc, w, h] format.
	"""
	# Convert nx4 boxes from [x1, y1, x2, y2] to [x, y, w, h] where xy1=top-left, xy2=bottom-right
	# x = np.array(x)
	y = torch.zeros_like(x) if isinstance(x, torch.Tensor) else np.zeros_like(x)
	if len(x.shape) == 1:
		y[0] = (x[0] + x[2]) / 2
		y[1] = (x[1] + x[3]) / 2
		y[2] = (x[2] - x[0]) / 2
		y[3] = (x[3] - x[1]) / 2
	elif len(x.shape) == 2:
		y[:, 0] = (x[:, 0] + x[:, 2]) / 2  # x center
		y[:, 1] = (x[:, 1] + x[:, 3]) / 2  # y center
		y[:, 2] = x[:, 2] - x[:, 0]  # width
		y[:, 3] = x[:, 3] - x[:, 1]  # height
	else:
		raise NotImplementedError(f'does not support dim: {len(x.shape)} shape: {x.shape}')
	return y


def xywh2xyxy(x, need_scale=False, im0=None):
	"""
	:param x: bbox in [xc, yc, w, h] format of torch.tensor or np.array.
	:return: bbox in [x1, y1, x2, y2] format
	"""
	# Convert nx4 boxes from [x, y, w, h] to [x1, y1, x2, y2] where xy1=top-left, xy2=bottom-right
	y = torch.zeros_like(x) if isinstance(x, torch.Tensor) else np.zeros_like(x)
	y[:, 0] = x[:, 0] - x[:, 2] / 2  # top left x
	y[:, 1] = x[:, 1] - x[:, 3] / 2  # top left y
	y[:, 2] = x[:, 0] + x[:, 2] / 2  # bottom right x
	y[:, 3] = x[:, 1] + x[:, 3] / 2  # bottom right y
	if need_scale:
		assert im0 is not None
		h, w = im0.shape[:2]
		y[:, 0] = y[:, 0] * w
		y[:, 1] = y[:, 1] * h
		y[:, 2] = y[:, 2] * w
		y[:, 3] = y[:, 3] * h
		y = y.astype(np.int)
	return y


def pts2bbox(pts):
	"""
	points to bbox
	:param pts: list or ndarray, [n, 2]
	:return: bbox in ndarray
	"""
	pts = np.array(pts)
	assert pts.ndim == 2
	bbox = np.array(
		[np.min(pts, axis=0)[0], np.min(pts, axis=0)[1],
		 np.max(pts, axis=0)[0], np.max(pts, axis=0)[1]])
	return bbox


def mask2bbox(mask):
	"""
	:param mask: mask, [w, h] or [w, h, 1]
	:return: bbox
	"""
	assert 2 <= mask.ndim <= 3
	if len(mask.shape) == 3:
		if mask.shape[2] != 1:
			raise NameError(f'mask shape error, {mask.shape}')
		mask = mask[:, :, 0]
	if mask.max() - mask.min() == 0:
		return None
	mask = np.array(mask / (mask.max() - mask.min()), dtype=np.uint8)  # 转成0到1
	# columns = np.argmax(mask, axis=0)
	pts = np.argwhere(mask == 1)  # 返回的是索引
	bbox = np.array([pts.min(axis=0)[1], pts.min(axis=0)[0], pts.max(axis=0)[1], pts.max(axis=0)[0]])
	# print(pts, pts.shape, bbox)
	return bbox



import functools
import numpy as np
import collections


def list2table(a, float_bit=2, alignment='<'):
	"""
	格式化，列表变表格，返回能直接打印的字符串。
	:param a: list, 长度为n，每个元素内由是一个str bool int float None list或dict
		example：
			[['COLUME1', 'COLUME2', 'COLUME2'], [1, True, 3.1415926]]]
	:param float_bit: 浮点数的小数位数
	:param alignment: 表格对齐方式
	:return: format str:
		example:
			COLUME1 COLUME2 COLUME3
			1       True    3.14
	"""
	assert isinstance(a, list)

	def convert(x):
		"""把所有内部元素转换为list"""
		f = functools.partial(isinstance, x)
		if f(str) or f(int) or f(bool):
			return [f'{x}']
		elif x is None:
			return ['None']
		elif f(float):
			return [f'{x:.{float_bit}f}']
		elif f(list):
			tmp = [convert(xx) for xx in x]
			return flatten_list(tmp)
		elif f(dict):
			return [f"Key('{xx}')" for xx in x.keys()]
		else:
			raise TypeError(f'Only support basic python types. value type: {type(x)}, x: {x}')

	def complection(x, max_lenth):
		"""补全内部list，用空补全"""
		tmp = ['-'] * max_lenth
		for i, xx in enumerate(x):
			tmp[i] = xx
		return tmp

	# 处理空值
	a = [convert(x) for x in a]
	max_lenth = np.max([len(x) for x in a])
	a = [complection(x, max_lenth=max_lenth) for x in a]

	# format string
	lenth = np.array([[len(x) for x in xx] for xx in a])  # (n, max_lenth)
	lenth = np.max(lenth, axis=0)  # (max_lenth,)
	format_str = [[f'{x:{alignment}{lenth[i]}}' for i, x in enumerate(xx)] for xx in a]
	format_str = '\n'.join(['  '.join(x) for x in format_str])
	return format_str


def flatten_list(a):
	"""展平list"""
	new_list = []
	[new_list.extend(x) for x in a]
	return new_list


def dict2info(info_dict):
	"""
	把一个字典的键和值展开为带缩进、带颜色的字符串，易于打印
	:param info_dict: 字典
	:return: str
	"""
	# 先递归地展开
	new_info_dict = collections.OrderedDict()
	indents = []
	indent_space = 4
	for k, v in info_dict.items():
		dict2info_recursive_func(k, v, new_info_dict, indents, indent_space=indent_space)
	info_dict = new_info_dict

	indents2color = {0: '32m', 1 * indent_space: '35m', 2 * indent_space: '36m', 3 * indent_space: '33m'}
	lenth = np.max([len(x) for x in info_dict.keys()])
	format_str = ''
	for i, (k, v) in enumerate(info_dict.items()):
		indent = indents[i]
		# print(f'k: {k} {len(k)}')
		k_str = f'{k.strip("*"):>{indent + lenth}}'
		color = indents2color.get(indent, None)
		k_str = f'\033[1;{color}{k_str}\033[0m' if color else k_str  # 上色
		format_str += f'  {k_str} : {v}\n'
	return format_str


def dict2info_recursive_func(k, v, new_info_dict, indents, indent=0, indent_space=4):
	"""递归拆分，原来的字典里，某些值可能还是dict，为了方便显示，递归展开，子dict的每个键作为新的顶层的键"""
	if isinstance(v, dict):
		while True:
			if k not in new_info_dict.keys():
				break
			k = '*' + k
		new_info_dict[f'{k}'] = f'({len(v.keys())})'
		indents.append(indent)
		indent += indent_space
		for k2, v2 in v.items():
			# print(f'k2: {k2} indent: {indent}', new_info_dict.keys())
			dict2info_recursive_func(k2, v2, new_info_dict, indents, indent, indent_space=indent_space)
	else:
		while True:
			if k not in new_info_dict.keys():
				break
			k = '*' + k
		new_info_dict[f'{k}'] = v
		indents.append(indent)
	return new_info_dict


if __name__ == '__main__':
	a = [[22, '1', ['xxx', 22]], 'x', {'a': 1, 'b': 3}]
	print(a[2].keys())
	ret = list2table(a)
	print(ret)



"""
Augment
"""
import os, sys
import numpy as np
import cv2
from copy import deepcopy

from ..functions.general import mask2bbox


class Augment(object):
	def __init__(self):
		"""
		注意：所有使用的mask, 形状是[h, w, 1]，宽高跟原始图相同，通道为1，取值范围二值化，即0和1
		"""
		pass

	def get_random_tzr_params(self, mask, existing_masks=[], thresh=0):
		"""
		传入二值化mask，输出随机尺寸、位置、缩放矩阵，如果同时传入存在的masks，保证遮挡率小于阈值
		:param mask: binary mask, ndarray [h, w, 1]
		:param existing_masks:
		:param thresh:
		:return:
		"""
		assert np.max(mask) == 1 and np.min(mask) == 0, f'mask must be binary, max: {np.max(mask)} min: {np.min(mask)}'
		mask = np.array(mask, dtype=np.uint8)
		h, w, c = mask.shape
		while True:
			# 随机尺寸
			size = np.clip(np.random.randn() / 6 * 2 + 1.3, 0.5, 2)
			angle = np.random.randint(0, 180)
			# 获取随机缩放和随机旋转矩阵
			matRotateZoom = self.get_rotation_zoom_matrix(mask, scale_factor=size, angle=angle)
			# 执行随机缩放和旋转
			new_mask = cv2.warpAffine(deepcopy(mask), matRotateZoom, (h, w))
			new_mask = new_mask[..., np.newaxis]
			# 随机位置
			new_mask, img, translation_params = self.random_position(new_mask, img)
			if new_mask is None:
				continue
			# b = [self.mask2bbox(new_mask)]
			# self.imshow(new_mask, bboxes=b, clss=[cls], show_name='new_mask')
			# self.imshow(img, bboxes=b, clss=[cls], show_name='img')
			# 判断交叉
			occ = [0]
			for j in range(len(existing_masks)):
				exist_mask = existing_masks[j]
				iop2 = self.cal_contains(mask1=new_mask, mask2=exist_mask)
				# intersection of percent 2, 新生成的mask遮挡已存在的mask的程度
				occ.append(iop2)
			if np.max(occ) <= thresh:
				# print(f'完全无遮挡: {occ}')
				return img, new_mask, matRotateZoom, translation_params

	def apply_transformation(self, ):
		# img = cv2.warpAffine(deepcopy(im0), matRotateZoom, (h, w))
		pass

	def translate_mask(self, mask, translation_params):
		x1, y1, x2, y2, bw, bh, stick_x, stick_y = translation_params
		new_mask = np.zeros_like(mask, dtype=np.uint8)  # [h, w, 1]
		cropped_bbox = mask[y1:y2, x1:x2, :]
		new_mask[stick_y:stick_y + bh, stick_x:stick_x + bw, :] = cropped_bbox
		return new_mask

	def get_rotation_zoom_matrix(self, mask, scale_factor=1, angle=0):
		"""
		根据给定的尺寸因子、旋转，返回旋转缩放矩阵
		:param mask: [h, w, c] c=1
		:param scale_factor: scale factor, default 1
		:param angle: rotation angle, default 0, rotation center: bbox center
		:return RotationMatrix2D
		"""
		sf = scale_factor
		bbox = mask2bbox(mask)
		xc, yc = bbox[1] + (bbox[3] - bbox[1]) / 2, bbox[0] + (bbox[2] - bbox[0]) / 2
		matRotateZoom = cv2.getRotationMatrix2D((xc, yc), angle, sf)
		return matRotateZoom

	def cal_contains(self, mask1, mask2):
		"""计算mask1和mask2中的目标的包含关系，即mask1与mask2的交集 除以 mask2的面积"""
		intersection = np.sum(mask2[mask1 == 1] == 1)  # 交集
		area2 = np.sum(mask2 == 1)
		correlation = intersection / (area2 + 1e-8)
		return correlation

	def random_position(self, mask, img):
		"""

		:param mask:
		:param img:
		:return:
		"""
		bbox = mask2bbox(mask)
		if bbox is None:
			return None, None, None
		h, w, c = mask.shape
		x1, y1, x2, y2 = bbox
		bw, bh = x2 - x1, y2 - y1  # bbox w h
		stick_x = np.random.randint(low=0, high=w - bw)
		stick_y = np.random.randint(low=0, high=h - bh)
		# translate mask
		translation_params = [x1, y1, x2, y2, bw, bh, stick_x, stick_y]
		new_mask = self.translate_mask(mask, translation_params)
		# translate img
		new_img = np.zeros((h, w, 3), dtype=np.uint8)
		new_img[stick_y:stick_y + bh, stick_x:stick_x + bw, :] = img[y1:y2, x1:x2, :]
		# print(f'maskshape: {mask.shape} sx: {stick_x} sy: {stick_y} bbox: {bbox} bw: {bw} bh: {bh}\n', cropped_bbox.shape, new_mask.shape)
		return new_mask, new_img, translation_params



"""
用pyinstaller把程序它打包成程序，源码封装成so,索引环境中的所有库，可以脱离配置环境运行。
"""

import os
import shutil
import damei as dm


class Packing(object):
	def __init__(self):
		self.app_name = 'mmlabelme_client'
		self.entrance = './mmlabelme/__main__.py'
		self.resources = []  # 可以为空
		self.resources = ['./mmlabelme/config', './mmlabelme/icons']

	def __call__(self, *args, **kwargs):
		app_name = self.app_name
		entrance = os.path.abspath(self.entrance)

		main_path = os.path.abspath('.')
		software_path = f'{main_path}/software'
		system = dm.current_system()
		software_path += f'/{system}'
		if not os.path.exists(software_path):
			os.makedirs(software_path)

		os.chdir(software_path)
		# os.system('rm -rf *')

		# 打包
		os.system(f'pyinstaller -n {app_name} {entrance}')

		# 删除build和spec文件
		os.system(f'rm -rf ./build')
		os.system(f'rm {app_name}.spec')

		# 创建app索引
		if system == 'linux':
			os.system(f'ln -s {software_path}/dist/{app_name}/{app_name} ./{app_name}')

		# 拷贝资源项
		for i, resource in enumerate(self.resources):
			print(f'\rCopy resources ... [{i + 1}/{len(self.resources)}]', end='')

			resource = resource[2::] if resource.startswith('./') else resource
			sp = f'{main_path}/{resource}'
			tp = f'{software_path}/dist/{app_name}/{resource}'
			if not os.path.exists(tp):
				# print(f'mkdir: {tp}')
				os.makedirs(tp)

			files = [x for x in os.listdir(sp) if not x.endswith('.py')]
			files = [x for x in files if os.path.isfile(f'{sp}/{x}')]
			# print(sp, tp, len(files))
			for j, file in enumerate(files):
				shutil.copy(f'{sp}/{file}', f'{tp}/{file}')


if __name__ == '__main__':
	p = Packing()
	p()



__name__ = "damei"

import subprocess

from damei import version

__version__ = version.__version__
__author__ = version.__author__

# dmsystem
from damei.dmsystem.system import current_system, system_lib_suffix

# controls
from damei.controls.color import ColorControl


def colors(num):
	return ColorControl(num).color


# functions
from damei.functions import general, torch_utils
from damei.functions.sub_process import popen

# post
from damei.post import post

# tools
from damei.tools.tools import Tools
# from damei.tools.tools import video2frames
tools = Tools()
from damei.tools.ffmpeg import DmFFMPEG
ffmpeg = DmFFMPEG()


from damei.tools.check_yolo import CheckYOLO

# misc
from damei.misc import misc
from damei.misc.logger import getLogger
from damei.misc.time import current_time, plus_time, within_time
from damei.misc.dm_rsa import DmRsa
rsa = DmRsa()
from damei.misc.config_loader import PyConfig
from damei.misc.scrcpy.scrcpy import Scrcpy





import subprocess


def popen(code, need_error=False):
	out = subprocess.Popen(code, shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
	if need_error:
		output = out.stdout.read().decode('utf-8').replace('\r', '').split('\n')
		error = out.stderr.read().decode('utf-8').replace('\r', '').split('\n')
		output = output[:-1] if output[-1] == '' else output
		error = error[:-1] if error[-1] == '' else error
		return output, error
	else:
		output = out.stdout.read().decode('utf-8').replace('\r', '').split('\n')
		output = output[:-1] if output[-1] == '' else output
		return output



import math
import os
import time
import logging
from copy import deepcopy

try:
	import torch
	import torch.backends.cudnn as cudnn
	import torch.nn as nn
	import torch.nn.functional as F
	import torchvision.models as models
except Exception as e:
	pass

logger = logging.getLogger(__name__)


def init_seeds(seed=0):
	torch.manual_seed(seed)

	# Speed-reproducibility tradeoff https://pytorch.org/docs/stable/notes/randomness.html
	if seed == 0:  # slower, more reproducible
		cudnn.deterministic = True
		cudnn.benchmark = False
	else:  # faster, less reproducible
		cudnn.deterministic = False
		cudnn.benchmark = True


def select_device(device='', batch_size=None):
	# device = 'cpu' or '0' or '0,1,2,3'
	cpu_request = device.lower() == 'cpu'
	if device and not cpu_request:  # if device requested other than 'cpu'
		os.environ['CUDA_VISIBLE_DEVICES'] = device  # set environment variable
		assert torch.cuda.is_available(), 'CUDA unavailable, invalid device %s requested' % device  # check availablity

	cuda = False if cpu_request else torch.cuda.is_available()
	if cuda:
		c = 1024 ** 2  # bytes to MB
		ng = torch.cuda.device_count()
		if ng > 1 and batch_size:  # check that batch_size is compatible with device_count
			assert batch_size % ng == 0, 'batch-size %g not multiple of GPU count %g' % (batch_size, ng)
		x = [torch.cuda.get_device_properties(i) for i in range(ng)]
		s = 'Using CUDA '
		for i in range(0, ng):
			if i == 1:
				s = ' ' * len(s)
			logger.info("%sdevice%g _CudaDeviceProperties(name='%s', total_memory=%dMB)" %
						(s, i, x[i].name, x[i].total_memory / c))
	else:
		logger.info('Using CPU')

	logger.info('')  # skip a line
	return torch.device('cuda:0' if cuda else 'cpu')


def time_synchronized():
	torch.cuda.synchronize() if torch.cuda.is_available() else None
	return time.time()


def is_parallel(model):
	return type(model) in (nn.parallel.DataParallel, nn.parallel.DistributedDataParallel)


def intersect_dicts(da, db, exclude=()):
	# Dictionary intersection of matching keys and shapes, omitting 'exclude' keys, using da values
	return {k: v for k, v in da.items() if k in db and not any(x in k for x in exclude) and v.shape == db[k].shape}


def initialize_weights(model):
	for m in model.modules():
		t = type(m)
		if t is nn.Conv2d:
			pass  # nn.init.kaiming_normal_(m.weight, mode='fan_out', nonlinearity='relu')
		elif t is nn.BatchNorm2d:
			m.eps = 1e-3
			m.momentum = 0.03
		elif t in [nn.LeakyReLU, nn.ReLU, nn.ReLU6]:
			m.inplace = True


def find_modules(model, mclass=None):
	mclass = mclass if mclass else nn.Conv2d
	# Finds layer indices matching module class 'mclass'
	return [i for i, m in enumerate(model.module_list) if isinstance(m, mclass)]


def sparsity(model):
	# Return global model sparsity
	a, b = 0., 0.
	for p in model.parameters():
		a += p.numel()
		b += (p == 0).sum()
	return b / a


def prune(model, amount=0.3):
	# Prune model to requested global sparsity
	import torch.nn.utils.prune as prune
	print('Pruning model... ', end='')
	for name, m in model.named_modules():
		if isinstance(m, nn.Conv2d):
			prune.l1_unstructured(m, name='weight', amount=amount)  # prune
			prune.remove(m, 'weight')  # make permanent
	print(' %.3g global sparsity' % sparsity(model))


def fuse_conv_and_bn(conv, bn):
	# https://tehnokv.com/posts/fusing-batchnorm-and-conv/
	with torch.no_grad():
		# init
		fusedconv = nn.Conv2d(conv.in_channels,
							  conv.out_channels,
							  kernel_size=conv.kernel_size,
							  stride=conv.stride,
							  padding=conv.padding,
							  bias=True).to(conv.weight.device)

		# prepare filters
		w_conv = conv.weight.clone().view(conv.out_channels, -1)
		w_bn = torch.diag(bn.weight.div(torch.sqrt(bn.eps + bn.running_var)))
		fusedconv.weight.copy_(torch.mm(w_bn, w_conv).view(fusedconv.weight.size()))

		# prepare spatial bias
		b_conv = torch.zeros(conv.weight.size(0), device=conv.weight.device) if conv.bias is None else conv.bias
		b_bn = bn.bias - bn.weight.mul(bn.running_mean).div(torch.sqrt(bn.running_var + bn.eps))
		fusedconv.bias.copy_(torch.mm(w_bn, b_conv.reshape(-1, 1)).reshape(-1) + b_bn)

		return fusedconv


def model_info(model, verbose=False):
	# Plots a line-by-line description of a PyTorch model
	n_p = sum(x.numel() for x in model.parameters())  # number parameters
	n_g = sum(x.numel() for x in model.parameters() if x.requires_grad)  # number gradients
	if verbose:
		print('%5s %40s %9s %12s %20s %10s %10s' % ('layer', 'name', 'gradient', 'parameters', 'shape', 'mu', 'sigma'))
		for i, (name, p) in enumerate(model.named_parameters()):
			name = name.replace('module_list.', '')
			print('%5g %40s %9s %12g %20s %10.3g %10.3g' %
				  (i, name, p.requires_grad, p.numel(), list(p.shape), p.mean(), p.std()))

	try:  # FLOPS
		from thop import profile
		flops = profile(deepcopy(model), inputs=(torch.zeros(1, 3, 64, 64),), verbose=False)[0] / 1E9 * 2
		fs = ', %.1f GFLOPS' % (flops * 100)  # 640x640 FLOPS
	except:
		fs = ''

	logger.info(
		'Model Summary: %g layers, %g parameters, %g gradients%s' % (len(list(model.parameters())), n_p, n_g, fs))


def load_classifier(name='resnet101', n=2):
	# Loads a pretrained model reshaped to n-class output
	model = models.__dict__[name](pretrained=True)

	# Display model properties
	input_size = [3, 224, 224]
	input_space = 'RGB'
	input_range = [0, 1]
	mean = [0.485, 0.456, 0.406]
	std = [0.229, 0.224, 0.225]
	for x in [input_size, input_space, input_range, mean, std]:
		print(x + ' =', eval(x))

	# Reshape output to n classes
	filters = model.fc.weight.shape[1]
	model.fc.bias = nn.Parameter(torch.zeros(n), requires_grad=True)
	model.fc.weight = nn.Parameter(torch.zeros(n, filters), requires_grad=True)
	model.fc.out_features = n
	return model


def scale_img(img, ratio=1.0, same_shape=False):  # img(16,3,256,416), r=ratio
	# scales img(bs,3,y,x) by ratio
	if ratio == 1.0:
		return img
	else:
		h, w = img.shape[2:]
		s = (int(h * ratio), int(w * ratio))  # new size
		img = F.interpolate(img, size=s, mode='bilinear', align_corners=False)  # resize
		if not same_shape:  # pad/crop img
			gs = 32  # (pixels) grid size
			h, w = [math.ceil(x * ratio / gs) * gs for x in (h, w)]
		return F.pad(img, [0, w - s[1], 0, h - s[0]], value=0.447)  # value = imagenet mean


def copy_attr(a, b, include=(), exclude=()):
	# Copy attributes from b to a, options to only include [...] and to exclude [...]
	for k, v in b.__dict__.items():
		if (len(include) and k not in include) or k.startswith('_') or k in exclude:
			continue
		else:
			setattr(a, k, v)


class ModelEMA:
	""" Model Exponential Moving Average from https://github.com/rwightman/pytorch-image-models
	Keep a moving average of everything in the model state_dict (parameters and buffers).
	This is intended to allow functionality like
	https://www.tensorflow.org/api_docs/python/tf/train/ExponentialMovingAverage
	A smoothed version of the weights is necessary for some training schemes to perform well.
	This class is sensitive where it is initialized in the sequence of model init,
	GPU assignment and distributed training wrappers.
	"""

	def __init__(self, model, decay=0.9999, updates=0):
		# Create EMA
		self.ema = deepcopy(model.module if is_parallel(model) else model).eval()  # FP32 EMA
		# if next(model.parameters()).device.type != 'cpu':
		#     self.ema.half()  # FP16 EMA
		self.updates = updates  # number of EMA updates
		self.decay = lambda x: decay * (1 - math.exp(-x / 2000))  # decay exponential ramp (to help early epochs)
		for p in self.ema.parameters():
			p.requires_grad_(False)

	def update(self, model):
		# Update EMA parameters
		with torch.no_grad():
			self.updates += 1
			d = self.decay(self.updates)

			msd = model.module.state_dict() if is_parallel(model) else model.state_dict()  # model state_dict
			for k, v in self.ema.state_dict().items():
				if v.dtype.is_floating_point:
					v *= d
					v += (1. - d) * msd[k].detach()

	def update_attr(self, model, include=(), exclude=('process_group', 'reducer')):
		# Update EMA attributes
		copy_attr(self.ema, model, include, exclude)



import os, sys


def current_system():
	"""
	get current system: linux, macos or windows
	:return:
	"""
	if sys.platform.startswith('linux'):
		return 'linux'
	elif sys.platform.startswith('darwin'):
		return 'macos'
	elif sys.platform.startswith('win32'):
		return 'windows'
	else:
		raise NameError(f'damei: current system is neither linux nor macos nor windows: {sys.platform}')


def system_lib_suffix():
	csystem = current_system()
	if csystem == 'linux':
		return '.so'
	elif csystem == 'macos':
		return '.dylib'
	elif csystem == 'windows':
		return '.dll'
	else:
		return None



import os, sys
from pathlib import Path
from easydict import EasyDict
import functools
import collections
import copy
import re
from . import misc


class PyConfig(collections.UserDict):
    def __init__(self, cfg_file=None, name=None, root_path=None):
        super(PyConfig, self).__init__()
        self.root_path = root_path
        # self.root_path = root_path if root_path else f'{pydir.parent.parent}'  # 上上级路径
        self._name = name if name else f'{cfg_file}'  # 其实是path
        self._items = dict()
        self.init_config(cfg_file)  # 根据配置文件把内容和属性注册到items里

        self.check_items()  # TODO

    def __len__(self):
        return len(self._items)

    def __dict__(self):
        return self._items

    def __getitem__(self, item):
        return self._items[item]

    def __repr__(self):
        format_str = f'<class {self.__class__.__name__}> ' \
                     f'(path="{self._name}", ' \
                     f'items={self.items}'
                     # f'keys={self._items.keys()})'

        return format_str

    def __missing__(self, key):
        if isinstance(key, str):
            raise KeyError(key)
        return self[str(key)]

    def __contains__(self, key):
        return str(key) in self.data

    def __setitem__(self, key, item):
        self.data[str(key)] = item

    # def __getattr__(self, key):
    #    return self.data[str(key)]

    def info(self):
        """查看当前所有配置信息"""
        # info_str = f'{self.__repr__()}\n'
        info_str = f'Configs:\n'
        info_dict = self.items
        info_str += misc.dict2info(info_dict)
        return info_str

    def init_config(self, cfg_file):
        if cfg_file is None:
            return
        cp = Path(os.path.abspath(cfg_file))  # config path
        # 根据sys.path的环境路径和cp的绝对路径，获取module_dir
        valid_sys_path = [x for x in sys.path if x in f'{cp}']
        if len(valid_sys_path) >= 1:
            rp = valid_sys_path[0]  # root_path
        else:
            sys.path.append(f'{cp.parent}')
            rp = f'{cp.parent}'
        # 提取模块路径：如：modules.detection.seyolov5.config
        module_dir = f'{str(cp.parent).replace(f"{rp}/", "").replace("/", ".")}'
        # print(module_dir)
        # print(f'rp: {rp} \ncp: {cp}\nm dir: {module_dir}')
        code = f'from {module_dir} import {cp.stem}'
        # print(cfg_file, code)
        exec(code)
        cfg = eval(cp.stem)  # 模块名

        # 读取文件获取顶格写并且不以#开头的属性
        with open(cfg_file, 'r') as f:
            data = f.readlines()
        data = [x for x in data if not (x.startswith(' ') or x.startswith('#'))]  # 非顶格的不要，# 开头的不要
        data = '\n'.join(data)
        attrs = re.findall(pattern=r"\s[a-z]\w+ ?= ?", string=data)
        attrs = [x.split('=')[0].replace('\n', '').strip() for x in attrs]

        # 注册属性到_items里，注册属性到self的属性里。
        for attr in attrs:
            if hasattr(cfg, attr):
                exec(f"self._items[attr] = cfg.{attr}")
                attr_value = self._items[attr]
                if isinstance(attr_value, dict):
                    setattr(self, attr, EasyDict(self._items[attr]))
                else:
                    setattr(self, attr, attr_value)
            else:
                pass

    def check_items(self):
        """初始化末，检查所有配置项，处理：缩略语~转为真实路径。
        递归方法，如果v类型是str,判断替换，如果循环完了，结束，如果v的类型是dict，继续循环
        """
        # print(self._items)
        for k, v in self._items.items():
            new_v = self.recursive_func(v)
            self._items[k] = new_v
            # 更新属性
            if hasattr(self, k):
                delattr(self, k)
            setattr(self, k, EasyDict(new_v)) if isinstance(new_v, dict) else setattr(self, k, new_v)

    def recursive_func(self, value):
        """递归函数,如果v类型是str,判断替换，如果循环完了，结束，如果v的类型是dict，继续循环"""
        if value is None:
            return None
        elif isinstance(value, str):
            return value.replace('~', os.environ['HOME'])
        elif isinstance(value, int):
            return value
        elif isinstance(value, float):
            return value
        elif isinstance(value, bool):
            return value
        elif isinstance(value, list):
            new_list = [self.recursive_func(x) for x in value]
            return new_list
        elif isinstance(value, dict):
            new_dict = {}
            for k, v in value.items():
                new_dict[k] = self.recursive_func(v)
            return new_dict
        else:
            raise TypeError(f'Config file only support basic python types but the value type is: {type(value)}.')

    def update_item(self, attrs, value):
        """更新配置条目
        attrs: 长度为0时代表，更新全部value，长度为1时，代表第1个元素是第一层，
        """
        if len(attrs) == 0:
            self.merge(value)
        elif len(attrs) == 1:
            self._items[attrs[0]] = value
        elif len(attrs) == 2:
            self._items[attrs[0]][attrs[1]] = value
        elif len(attrs) == 3:
            self._items[attrs[0]][attrs[1]][attrs[2]] = value
        else:
            self._items[attrs[0]][attrs[1]][attrs[2]][attrs[3]] = value
        self.check_items()

    def merge(self, cfg2):
        """合并另一个配置文件到内部，是inplace的"""
        if cfg2 is None:
            return self
        for k2, v2 in cfg2._items.items():
            merged_v = self.recursive_func2(v2, k2=k2, sub_items=self._items)
            self._items[k2] = merged_v
        # print(self.items)
        self.check_items()
        return self

    def recursive_func2(self, v2, k2=None, sub_items=None):
        """v1是之前的值，v2是现在的值"""
        if k2 and (k2 not in list(sub_items.keys())):
            return v2
        else:
            v1 = sub_items[k2]

            if type(v1) != type(v2):
                return v2
            f = functools.partial(isinstance, v1)
            if f(str) or f(int) or f(float) or f(bool):
                return v2
            elif f(list):
                tmp = [x for x in v2 if x not in v1]
                return v1 + tmp
            elif f(dict):  # v1v2都是dict
                new_dict = v1  # 保留v1
                for kk, vv in v2.items():
                    new_dict[kk] = self.recursive_func2(vv, k2=kk, sub_items=v1)
                return new_dict

    @property
    def items(self):
        return self._items

    def keys(self):
        return self._items.keys()

    def get(self, key, default=None):
        if key in list(self._items.keys()):
            return self._items[key]
        else:
            return default










"""
利用ffmpeg推流
"""
import os
import numpy as np
import subprocess
from damei.misc.logger import getLogger

logger = getLogger('DmFFMPEG')


class DmFFMPEG(object):
	def __init__(
			self,
			ip='127.0.0.1',
			port='1935',
			stream_type='rtmp',
			key=None,
			vcodec='h264',
			acodec=None,
			**kwargs
		):
		self.ip = ip
		self.port = port
		self.stream_type = stream_type
		self.key = key
		self.vcodec = vcodec
		self.acodec = acodec
		self._pipe = None
		self.mute = kwargs.pop('mute', False)  #

	def init_pipe(self, img, **kwargs):
		"""初始化管道，第一次调用push_stream时且传入为图像时，调用"""
		ip = kwargs.pop('ip', self.ip)
		port = kwargs.pop('port', self.port)
		stream_type = kwargs.pop('stream_type', self.stream_type)
		key = kwargs.pop('key', self.key)
		vcodec = kwargs.pop('vcodec', self.vcodec)
		acodec = kwargs.pop('acodec', self.acodec)
		fps = kwargs.pop('fps', 15)

		key = f'/{key}' if key else ''
		vcodec = f'{vcodec}' if vcodec else ''
		acodec = f'{acodec}' if acodec else ''

		url = f'{stream_type}://{ip}:{port}/live{key}'
		h, w, c = img.shape
		command = [
			'ffmpeg',
			# 're',#
			# '-y', # 无需询问即可覆盖输出文件
			'-f', 'rawvideo',  # 强制输入或输出文件格式
			# '-vcodec','rawvideo', # 设置视频编解码器。这是-codec:v的别名
			'-pix_fmt', 'bgr24',  # 设置像素格式
			'-s', f'{w}x{h}',  # 设置图像大小
			'-r', f'{fps}',  # 设置帧率
			'-i', '-',  # 输入
			'-vcodec', vcodec,
			'-acodec', acodec,
			# '-pix_fmt', 'yuv420p',
			# '-preset', 'ultrafast',
			'-f', 'flv',  # 强制输入或输出文件格式
			url]
		# print(command)
		# exit()
		logger.info(f'The stream will be pushed to: {url}.')

		if self.mute:
			pipe = subprocess.Popen(
				command, stdin=subprocess.PIPE, stderr=subprocess.STDOUT,
				stdout=subprocess.PIPE)
		else:
			pipe = subprocess.Popen(
				command, stdin=subprocess.PIPE,
			)
		# exit()
		self._pipe = pipe
		return pipe

	def get_pipe(self, img=None, **kwargs):
		if self._pipe is not None:
			return self._pipe
		else:
			return self.init_pipe(img=img, **kwargs)
			# return self._pipe

	def push_stream(self, source, **kwargs):
		"""
		推流
		:param source: 可以是rtsp rtmp流或.mp4 .avi等视频文件
						也可以是单张图，必须是np.ndarray格式
		:param kwargs:
		:return:
		"""
		ip = kwargs.pop('ip', self.ip)
		port = kwargs.pop('port', self.port)
		stream_type = kwargs.pop('stream_type', self.stream_type)
		key = kwargs.pop('key', self.stream_type)
		vcodec = kwargs.pop('vcodec', self.vcodec)
		acodec = kwargs.pop('acodec', self.acodec)
		key = f'/{key}' if key else ''
		vcodec = f'-vcodec {vcodec}' if vcodec else ''
		acodec = f'-acodec {acodec}' if acodec else ''

		# 推视频流
		if isinstance(source, str):
			code = f'ffmpeg -i {source} {vcodec} {acodec} -f flv {stream_type}://{ip}:{port}/live{key}'
			if self.mute:
				subprocess.Popen(code, stdout=subprocess.STDOUT, stderr=subprocess.STDOUT)
			else:
				os.system(code)
		elif isinstance(source, np.ndarray):
			# 第一次调用时检查是否有输出pipe
			pipe = self.get_pipe(source, **kwargs)
			pipe.stdin.write(source.tobytes())
		else:
			raise NotImplementedError(f'Source type error: {source} {type(source)}.')


	def show_stream(self, source=None, stream_type='rtmp', ip='127.0.0.1', port='1935', key=None):
		key = f'/{key}' if key else ''
		source = source if source else f'{stream_type}://{ip}:{port}/live{key}'
		code = f'ffplay {source}'
		os.system(code)








"""
检查YOLO数据集
"""
import os
from pathlib import Path
import numpy as np
import cv2
from ..functions import general
from ..controls.color import ColorControl


class CheckYOLO(object):
	def __init__(self, dp=None):
		self.dp = dp
		self.suffixes = ['.jpg', '.png', '.bmp']

	def __call__(self, trte=None):
		dp = self.dp
		trte = trte if trte else 'train'
		print(dp)
		self.classes = self.get_classes()

		num_color = len(self.classes) if self.classes is not None else 1000
		self.colors = ColorControl(num=num_color, random_color=True).color
		self.check(trte=trte)

	def get_classes(self):
		dp = self.dp
		if os.path.exists(f'{dp}/classes.txt'):
			with open(f'{dp}/classes.txt', 'r') as f:
				classes = f.readlines()
			classes = [x.replace('\n', '') for x in classes]
		else:
			classes = None
		return classes

	def check(self, trte=None, p=None, pl=None, names=None):
		p = p if p else f'{self.dp}/images/{trte}'
		pl = pl if pl else f'{self.dp}/labels/{trte}'
		names = names if names else self.classes
		if os.path.exists(p):
			pass
		else:
			print(f'Directory {p} does not exists, redirect to {self.dp}')
			p = self.dp
			pl = self.dp

		imgs = [f'{p}/{x}' for x in os.listdir(p) if Path(x).suffix in self.suffixes]
		imgs = sorted(imgs)
		# print(imgs)
		print(f'all classes: {names} len {len(imgs)}')
		for i, imgp in enumerate(imgs):
			stem = Path(imgp).stem
			labelp = f'{pl}/{Path(imgp).stem}.txt'
			img = cv2.imread(imgp)
			h, w, c = img.shape

			with open(labelp, 'r') as f:
				label = f.readlines()
			label = np.array([x.split() for x in label], dtype=np.float32)

			classes = label[:, 0]
			bboxes = label[:, 1::]
			bboxes = general.xywh2xyxy(bboxes)
			lbs = []
			for j in range(len(label)):
				cls = classes[j]
				bbox = bboxes[j]
				bbox[0] *= w
				bbox[1] *= h
				bbox[2] *= w
				bbox[3] *= h

				label = f'{names[int(cls)]}' if names is not None else f'{cls}'
				lbs.append(label)
				color = None

				general.plot_one_box_trace_pose_status(
					bbox, img, label=label, color=color)

			print(f'stem: {stem} img shape: {img.shape} num_targets: {len(classes)} lbs: {lbs}')
			# cr = np.any(label[:, 0] == 1)
			cr = True
			if cr:
				if np.max(img.shape) > 1920:
					h, w, c = img.shape
					img = cv2.resize(img, (int(0.5 * w), int(0.5 * h)))
				cv2.imshow('xx', img)
				if cv2.waitKey(0) == ord('q'):
					exit()



"""
转换训练集，YOLOv5 format to COCO format
"""
import os, sys
import argparse
from pathlib import Path
import json
import shutil
import cv2
from tqdm import tqdm


class YOLO2COCO(object):
	def __init__(self, sp, tp=None):
		self.sp = sp  # source path
		self.tp = tp if tp is not None else f'{Path(sp).parent}/transformed_coco_format'

		# 检查输出文件夹
		if os.path.exists(self.tp):
			ipt = input(f'Target path: {self.tp} exists, \nremove and continue? [YES/no]: ')
			if ipt in ['Y', '', 'y', 'yes', 'YES']:
				shutil.rmtree(self.tp)
			else:
				print('Exit')
				exit()
		os.makedirs(self.tp)
		os.makedirs(f'{self.tp}/annotations')
		os.makedirs(f'{self.tp}/train2017')
		os.makedirs(f'{self.tp}/val2017')

		# 关于coco格式
		self.annotation_id = 1
		self.type = 'instances'
		self.categories = self.read_categories()
		self.info = {
			'year': 2021,
			'version': '1.0',
			'description': 'For object detection',
			'date_created': '2021-9-28',
		}
		self.licenses = [{
			'id': 1,
			'name': 'GNU General Public License v3.0',
			'url': '',
		}]

		# 支持的图像格式
		self.suffix = ['.jpg', '.bmp', '.png']

	def read_categories(self):
		file = f'{self.sp}/classes.txt'
		print(self.sp)
		assert os.path.exists(file), f'classes file {file} does not exists.'
		with open(file, 'r') as f:
			data = f.readlines()
		data = [x.replace('\n', '') for x in data]
		categories = []
		for i, cls_name in enumerate(data):
			cls = i + 1
			categories.append({
				'supercategory': cls_name,
				'id': cls,
				'name': cls_name
			})
		return categories

	def __call__(self, *args, **kwargs):
		sp = self.sp
		for trte in ['train', 'tests']:
			print(f'Deal with {trte}')
			imgs = os.listdir(f'{sp}/images/{trte}')
			imgs = [f'{sp}/images/{trte}/{x}' for x in imgs if str(Path(x).suffix) in self.suffix]

			trval = trte if trte == 'train' else 'val'
			self.deal_single(img_paths=imgs, trval=trval)

	def deal_single(self, img_paths, trval):
		tp = self.tp
		bar = tqdm(img_paths)

		images = []
		annotations = []
		for i, img_path in enumerate(bar):
			img_id = i + 1
			stem = Path(img_path).stem
			txt_path = f'{str(Path(img_path).parent).replace("images", "labels")}/{stem}.txt'

			img = cv2.imread(img_path)
			h, w, c = img.shape

			new_stem = f'{img_id:0>12}'

			# 保存图像
			if Path(img_path).suffix.lower() == '.jpg':
				shutil.copyfile(img_path, f'{tp}/{trval}2017/{new_stem}.jpg')
			else:
				cv2.imwrite(f'{tp}/{trval}2017/{new_stem}.jpg')

			# 保存标注
			images.append({
				'date_captured': '2021',
				'file_name': f'{new_stem}.jpg',
				'id': img_id,
				'height': h,
				'width': w,
			})

			annot = self.label2annot(txt_path, h, w, img_id=img_id)
			assert len(annot) > 0, f'{txt_path} is empty.'
			annotations.extend(annot)

		json_data = {
			'info': self.info,
			'images': images,
			'licenses': self.licenses,
			'type': self.type,
			'annotations': annotations,
			'categories': self.categories
		}

		tp_json = f'{tp}/annotations/instances_{trval}2017.json'
		with open(tp_json, 'w', encoding='utf-8') as f:
			json.dump(json_data, f, ensure_ascii=False)

	def label2annot(self, txt_path, h, w, img_id):
		annotation = []
		with open(txt_path, 'r') as f:
			labels = f.readlines()
		labels = [x.replace('\n', '') for x in labels]
		for lb in labels:
			cls = lb.split()[0]
			bbox = lb.split()[1::]  # xc yc w h in fraction
			assert len(bbox) == 4
			segmentation, bbox, area = self._get_annotation(bbox, h, w)  # 转为x1y1wh了
			annotation.append({
				'segmentation': segmentation,
				'area': area,
				'iscrowd': 0,
				'image_id': img_id,
				'bbox': bbox,
				'category_id': int(cls) + 1,
				'id': self.annotation_id,
			})
			self.annotation_id += 1
		return annotation

	@staticmethod
	def _get_annotation(vertex_info, height, width):
		cx, cy, w, h = [float(i) for i in vertex_info]

		cx = cx * width
		cy = cy * height
		box_w = w * width
		box_h = h * height

		# left top
		x0 = max(cx - box_w / 2, 0)
		y0 = max(cy - box_h / 2, 0)

		# right bottomt
		x1 = min(x0 + box_w, width)
		y1 = min(y0 + box_h, height)

		segmentation = [[x0, y0, x1, y0, x1, y1, x0, y1]]
		bbox = [x0, y0, box_w, box_h]
		area = box_w * box_h
		return segmentation, bbox, area


if __name__ == '__main__':
	sp = '/home/zzd/datasets/crosswalk/fogged_train_data_v5_format'
	tp = "/home/zzd/datasets/crosswalk/fogged_train_data_coco_format"
	y2c = YOLO2COCO(sp, tp)
	y2c()



"""
post process
"""
from pathlib import Path
import struct
from damei.functions import torch_utils


def pt2wts(weights_path, output_dir=None):
	"""
	convert pytorch trained model weights [.pt] to tensorrt pre-weights [.wts]
	:param weights_path: /path/to/xxx.pt
	:param output_dir: directory for output xxx.wts, if None, the same as weights_dir.
	:return: None, generate xxx.wts weights in output_dir.
	"""
	import torch

	output_dir = Path(weights_path).parent if output_dir is None else output_dir
	stem = Path(weights_path).stem
	output_path = f'{output_dir}/{stem}.wts'
	# Initialize
	device = torch_utils.select_device('cpu')
	# Load models
	model = torch.load(weights_path, map_location=device)['model'].float()  # load to FP32
	model.to(device).eval()

	f = open(output_path, 'w')
	f.write('{}\n'.format(len(model.state_dict().keys())))
	print('saving...')
	for k, v in model.state_dict().items():
		vr = v.reshape(-1).cpu().numpy()
		f.write('{} {} '.format(k, len(vr)))
		for vv in vr:
			f.write(' ')
			f.write(struct.pack('>f', float(vv)).hex())
		f.write('\n')
	f.close()
	print(f'{stem}.wts save to {output_path}.')



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Note: To use the 'upload' functionality of this file, you must:
#   $ pipenv install twine --dev

import io
import os
import sys
from shutil import rmtree

from setuptools import find_packages, setup, Command

# Package meta-data.
NAME = 'damei'
DESCRIPTION = 'Da Mei project with deep learning general functions.'
URL = 'https://github.com/zhangzhengde0225/damei'
EMAIL = 'drivener@163.com'
AUTHOR = 'Zhengde Zhang'
REQUIRES_PYTHON = '>=3.6.0'

with open('damei/version.py') as f:
	data = f.readlines()
VERSION = eval(data[0].split('=')[-1].strip())
print(f'Installing damei, version: {VERSION}')

# What packages are required for this module to be executed?
REQUIRED = [
	# 'requests', 'maya', 'records',
]

# What packages are optional?
EXTRAS = {
	# 'fancy feature': ['django'],
}

# The rest you shouldn't have to touch too much :)
# ------------------------------------------------
# Except, perhaps the License and Trove Classifiers!
# If you do change the License, remember to change the Trove Classifier for that!

here = os.path.abspath(os.path.dirname(__file__))

# Import the README and use it as the long-description.
# Note: this will only work if 'README.md' is present in your MANIFEST.in file!
try:
	with io.open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
		long_description = '\n' + f.read()
except FileNotFoundError:
	long_description = DESCRIPTION

# Load the package's __version__.py module as a dictionary.
about = {}
if not VERSION:
	project_slug = NAME.lower().replace("-", "_").replace(" ", "_")
	with open(os.path.join(here, project_slug, '__version__.py')) as f:
		exec(f.read(), about)
else:
	about['__version__'] = VERSION


class UploadCommand(Command):
	"""Support setup.py upload."""

	description = 'Build and publish the package.'
	user_options = []

	@staticmethod
	def status(s):
		"""Prints things in bold."""
		print('\033[1m{0}\033[0m'.format(s))

	def initialize_options(self):
		pass

	def finalize_options(self):
		pass

	def run(self):
		try:
			self.status('Removing previous builds…')
			rmtree(os.path.join(here, 'dist'))
		except OSError:
			pass

		self.status('Building Source and Wheel (universal) distribution…')
		os.system('{0} setup.py sdist bdist_wheel --universal'.format(sys.executable))

		self.status('Uploading the package to PyPI via Twine…')
		try:
			import twine  # 需要用twine上传，但twine不一定安装，如果没装报错
		except:
			raise NameError(
				f'You need twine to upload the package to pypi.\n'
				f'           Install twine with "pip install twine" or switch a environment with twine.')
		os.system('twine upload dist/*')

		self.status('Pushing git tags…')
		os.system('git tag v{0}'.format(about['__version__']))
		os.system('git push --tags')

		sys.exit()


# Where the magic happens:
setup(
	name=NAME,
	version=about['__version__'],
	description=DESCRIPTION,
	long_description=long_description,
	long_description_content_type='text/markdown',
	author=AUTHOR,
	author_email=EMAIL,
	python_requires=REQUIRES_PYTHON,
	url=URL,
	packages=find_packages(exclude=["tests", "*.tests", "*.tests.*", "tests.*"]),
	# If your package is a single module, use this instead of 'packages':
	# py_modules=['mypackage'],

	# entry_points={
	#     'console_scripts': ['mycli=mymodule:cli'],
	# },
	install_requires=REQUIRED,
	extras_require=EXTRAS,
	include_package_data=True,
	license='MIT',
	classifiers=[
		# Trove classifiers
		# Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
		'License :: OSI Approved :: MIT License',
		'Programming Language :: Python',
		'Programming Language :: Python :: 3',
		'Programming Language :: Python :: 3.6',
		'Programming Language :: Python :: Implementation :: CPython',
		'Programming Language :: Python :: Implementation :: PyPy'
	],
	# $ setup.py publish support.
	cmdclass={
		'upload': UploadCommand,
	},
)



import os
import time
from pathlib import Path
import logging


def getLogger(name=None, **kwargs):
	name = name if name else 'root'
	name_lenth = kwargs.get('name_lenth', 12)
	name = f'{name:<{name_lenth}}'
	logger = logging.getLogger(name)
	level = kwargs.get('level', logging.INFO)
	format_str = f"\033[1;35m[%(asctime)s]\033[0m \033[1;32m[%(name)s]\033[0m " \
				 f"\033[1;36m[%(levelname)s]:\033[0m %(message)s"
	logging.basicConfig(level=level,
						format=format_str,
						# datefmt='%d %b %Y %H:%M:%S'
						)
	logg_dir = f'/tmp/{Path(os.getcwd()).name}'
	if not os.path.exists(logg_dir):
		os.makedirs(logg_dir)
	fh = logging.FileHandler(f'{logg_dir}/{Path(os.getcwd()).name}.log')
	fh.setLevel(level=level)
	# ch = logging.StreamHandler()
	# ch.setLevel(level=level)
	fh.setFormatter(logging.Formatter(format_str))
	# ch.setFormatter(logging.Formatter(format_str))
	logger.addHandler(fh)
	# logger.addHandler(ch)
	return logger



import rsa as raw_rsa
import base64


class DmRsa(object):
	"""这是加密Encrypt方案"""

	def __init__(self, plan='Encrypt'):
		"""
		:param plan: 加密方案Encrypt或签密方案SignEncripy
		"""
		self.length = 1024

	def gen_rsa(self, num=1, length=None):
		"""在随机生成公钥-私钥对，存储在当前路径pubk.txt和privk.txt里"""
		length = length if length else self.length
		# rsa_length = int(length / 8)
		# enable_length = rsa_length - 11
		for i in range(num):
			pubk, privk = self.gen_pair(length=length)
			self._save(pubk=pubk, privk=privk, suffix=f'{i + 1}')

	def gen_pair(self, length):
		pub_key_obj, priv_key_obj = raw_rsa.newkeys(length)
		pub_key_str = pub_key_obj.save_pkcs1()
		pub_key_code = base64.standard_b64encode(pub_key_str).decode()

		priv_key_str = priv_key_obj.save_pkcs1()
		priv_key_code = base64.standard_b64encode(priv_key_str).decode()
		return pub_key_code, priv_key_code

	def _save(self, pubk, privk, suffix=''):
		with open(f'pubk{suffix}.pem', 'w') as f:
			f.write(pubk)
		with open(f'privk{suffix}.pem', 'w') as f:
			f.write(privk)

	def load_pk(self, pubk=None, privk=None):
		"""加载钥匙"""
		assert pubk or privk  # 至少有一个
		assert not (pubk and privk)  # 不能同时都有

		if (pubk is not None) and (privk is not None):
			# print(privk, pubk, pubk is not None, privk is not None)
			raise KeyError(f'The input public key and private key both not None, the software dont know which to load.')
		elif pubk:
			if pubk.endswith('.pem'):
				with open(pubk, 'r') as f:
					pubk = f.read()
			pubk = base64.standard_b64decode(pubk)  # 对公钥解密
			pubk = raw_rsa.PublicKey.load_pkcs1(pubk)  # 加载公钥
			# print(f'pubk2: {pubk}')
			return pubk
		elif privk:
			if privk.endswith('.pem'):
				with open(privk, 'r') as f:
					privk = f.read()
			privk = base64.standard_b64decode(privk)  # 私钥解密
			privk = raw_rsa.PrivateKey.load_pkcs1(privk)  # 加载私钥
			return privk
		else:
			raise KeyError(f'Both public key and private key are None.')

	def encrypt(self, plaintext, pubk=None, privk=None, length=None):
		"""
		用公钥或私钥对明文进行rsa加密 + base64加密
		:param plaintext，明文，str类型
		:param pubk: 公钥，str类型，需先解密，直接是公钥，或者是以.txt结尾的公钥文件
		:return: 密文，str类型
		"""
		length = length if length else self.length
		# rsa加密
		pk = self.load_pk(pubk=pubk, privk=privk)
		rsa_length = int(length / 8)
		enable_length = rsa_length - 11
		ciphertext_list = []
		for i in range(0, len(plaintext), enable_length):
			v = plaintext[i:i + enable_length]
			val = raw_rsa.encrypt(v.encode(), pk)  # 加密，加密后是bytes类型
			ciphertext_list.append(val)
		ciphertext = b''.join(ciphertext_list)

		# base64加密
		ciphertext = base64.b64encode(ciphertext)
		ciphertext = ciphertext.decode('utf-8')

		return ciphertext

	def decrypt(self, ciphertext, privk=None, pubk=None, length=None):
		"""
		利用私钥或公钥，对密文进行解密，对应的rsa+base64加密，解密是base64+rsa解密
		:param ciphertext: 密文，str类型
		:param privk: 私钥，str类型
		:param length: 1024
		:return: 明文，str类
		"""
		# 先base64解密
		ciphertext = base64.b64decode(ciphertext.encode())

		# rsa解密
		length = length if length else self.length
		pk = self.load_pk(pubk=pubk, privk=privk)
		# print(f'pk: {pk} {type(pk)}')
		rsa_length = int(length / 8)
		plaintext_list = []
		for i in range(0, len(ciphertext), rsa_length):
			v = ciphertext[i:i + rsa_length]
			val = raw_rsa.decrypt(v, pk)
			plaintext_list.append(val)

		plaintext = b''.join(plaintext_list)
		plaintext = plaintext.decode('utf-8')
		return plaintext

		pass

	def sign(self, message, privk, hash_method='MD5'):
		"""
		Sign a message by private key
		:param message: 消息
		:param privk: 私钥
		:param hash_method: MD5, SHA-1, SHA-256, SHA-512等
		:return: 签名，16进制的str
		"""
		privk = self.load_pk(privk=privk)
		signature = raw_rsa.sign(message.encode('utf-8'), privk, hash_method=hash_method)
		# print('xxx', signature)
		signature = signature.hex()
		return signature

	def verify(self, message, signature, pubk):
		pubk = self.load_pk(pubk)
		# print('xx2', signature)
		# signature = signature.encode('utf-8')
		signature = bytes.fromhex(signature)
		# print('xx3', signature)
		try:
			raw_rsa.verify(message.encode('utf-8'), signature, pubk)
			return True
		except Exception as e:
			print(f'Verify failed, {e}')
			return False
	# print(ret)



import random
import collections


class ColorControl(object):
	def __init__(self, num=None, filt_name=None, random_color=True, rgb=None):
		self.color = self.get_color(
			num=num, filt_name=filt_name, random_color=random_color, rgb=rgb)

	def get_colors_from_inner(self, filt_name=None, rgb=None):
		inner_colors = collections.OrderedDict({
			'dark_blue': ["#003399", (0, 51, 153)],
			'dark_red': ["#F00000", (240, 0, 0)],
			'dark_yellow': ["#CCCC00", (204, 204, 0)],
			'dark_green': ["#009966", (0, 153, 102)],
			'dark_purple': ["#663366", (102, 51, 102)],
			'dark_orange': ["#FF6600", (255, 102, 0)],
			'light_blue:': ["#CCFFFF", (204, 255, 255)],
			'light_red': ["#FF6666", (255, 102, 102)],
			'light_black': ["#CCCCCC", (204, 204, 204)],
			'light_green': ['#CCFF99', (204, 255, 153)],
			'light_pink': ['#FFCCCC', (255, 204, 204)],
			'light_orange': ["#FF9966", (255, 153, 102)]
		})
		# filter
		if filt_name is None:
			colors = list(inner_colors.values())
		else:
			colors = [v for k, v in inner_colors if filt_name in k]
		assert len(colors) >= 1
		if rgb is None:  # 取字符串
			colors = [x[0] for x in colors]
		elif rgb == 'RGB':
			colors = [x[1] for x in colors]
		elif rgb == 'BGR':
			# print(colors[0])
			colors = [(x[1][2], x[1][1], x[1][0]) for x in colors]
		return colors

	def get_color(self, num=None, filt_name=None, random_color=True, rgb=None):
		"""
		方案：
		1.如果随机，直接随机取num个颜色，num为None时取1000个
		2.不随机，根据filt_name确定内置颜色，颜色名字中包含filt_name的颜色被选中。
			同时，不设置num, 全部取回，设置了num，内置颜色大于num, 取回num个，内置颜色小于None, 后面的用random补全，
		:param num:
		:param filt_name: None, 取所有内置颜色
		:return:
		"""
		if random_color:
			num = 1000 if num is None else num
			colors = [[random.randint(0, 255) for _ in range(3)] for __ in range(num)]
		else:
			colors = self.get_colors_from_inner(filt_name=filt_name, rgb=rgb)
			if num is not None:
				# assert len(colors) >= num, f'Number of filtered inner color: {len(colors)} less than num para: {num}.'
				if len(colors) < num:  # 补全
					rand_colors = [[random.randint(0, 255) for _ in range(3)] for __ in range(num - len(colors))]
					colors = colors + rand_colors
				if len(colors) == num:
					pass
				else:  # > num
					colors = colors[:num]
		return colors
